int
    TEXTBUFFERLENGTH = 256,		/* size of text buffers */
    INDENTAMOUNT = 4,			/* size of each displayed indent */
    MAXEDITDEPTH = 20;			/* maximum nesting when editing */

byte CONTROL = 0x1f;

type
    KINDTYPE = enum {
	KCONSTANT,
	KLOCALVAR,
	KGLOBALVAR,
	KPARAMETER,
	KPROC
    },

    TYPETYPE = enum {
	NULLTYPE,
	VOIDTYPE,
	BOOLTYPE,
	INTTYPE,
	STRINGTYPE
    },

    SYMBOLTYPE = struct {
	KINDTYPE sy_kind;
	TYPETYPE sy_type;
	union {
	    *PROCTYPE procValue;
	    *char stringValue;
	    int intValue;
	    bool boolValue;
	} sy_;
	*char sy_name;
    },

    UNARYOPTYPE = enum {
	OPBOOLNOT,
	OPNEGATE,
	OPABSOLUTE,
	OPLENGTH
    },

    BINARYOPTYPE = enum {
	OPBOOLAND,
	OPBOOLOR,
	OPPLUS,
	OPMINUS,
	OPTIMES,
	OPDIVIDE,
	OPMODULO,
	OPCONCATENATE
    },

    NODETYPETYPE = enum {
	IFNODE,
	WHILENODE,
	FORNODE,
	ASSIGNNODE,
	CALLNODE,
	READNODE,
	WRITENODE,
	UNARYNODE,
	BINARYNODE,
	BOOLNODE,
	NUMBERNODE,
	STRINGNODE
    },

    IDLISTTYPE = struct {
	*IDLISTTYPE id_next;
	*SYMBOLTYPE id_this;
    },

    DECLLISTTYPE = struct {
	*DECLLISTTYPE dc_next;
	TYPETYPE dc_type;
	*IDLISTTYPE dc_this;
    },

    NODELISTTYPE = struct {
	*NODELISTTYPE nl_next;
	*NODETYPE nl_this;
    },

    PROCTYPE = struct {
	TYPETYPE p_resultType;
	*IDLISTTYPE p_symbolTable;
	*DECLLISTTYPE p_parameters, p_locals;
	*NODELISTTYPE p_body;
	*NODETYPE p_result;
	*SYMBOLTYPE p_symbol;
    },

    IFLISTTYPE = struct {
	*IFLISTTYPE if_next;
	*NODETYPE if_condition;
	*NODELISTTYPE if_body;
    },

    IFTYPE = struct {
	*IFLISTTYPE if_alternates;
	*NODELISTTYPE if_elsePart;
    },

    WHILETYPE = struct {
	*NODELISTTYPE wh_condition, wh_body;
    },

    FORTYPE = struct {
	*NODETYPE f_variable, f_from, f_by, f_to;
	bool f_isUpto;
	*NODELISTTYPE f_body;
    },

    ASSIGNTYPE = struct {
	*NODETYPE a_destination, a_source;
    },

    CALLTYPE = struct {
	*SYMBOLTYPE c_proc;
	*NODELISTTYPE c_parameters;
    },

    READWRITETYPE = struct {
	*NODELISTTYPE rw_expressions;
    },

    UNARYTYPE = struct {
	UNARYOPTYPE u_operator;
	*NODETYPE u_operand;
    },

    BINARYTYPE = struct {
	BINARYOPTYPE b_operator;
	*NODETYPE b_left, b_right;
    },

    NODETYPE = struct {
	NODETYPETYPE n_type;
	union {
	    IFTYPE n_if;
	    WHILETYPE n_while;
	    FORTYPE n_for;
	    ASSIGNTYPE n_assign;
	    CALLTYPE n_call;
	    READWRITETYPE n_readWrite;
	    UNARYTYPE n_unary;
	    BINARYTYPE n_binary;
	    bool n_bool;
	    int n_number;
	    *char n_string;
	} n_;
    },

    P_TYPETYPE = enum {
	P_PROC,
	P_DECLLIST,
	P_IDLIST,
	P_NODELIST,
	P_IFLIST,
    },

    P_POSITIONTYPE = union {
	enum {
	    D_TYPE,
	    D_IDLIST
	} p_declaration;
	enum {
	    P_NAME,
	    P_PARAMETERS,
	    P_RESULTTYPE,
	    P_LOCALS,
	    P_BODY,
	    P_RESULT
	} p_proc;
	enum {
	    A_DESTINATION,
	    A_SOURCE
	} p_assignment;
	enum {
	    I_CONDITION,
	    I_ALTERNATIVE,
	    I_ELSE
	} p_if;
	enum {
	    W_CONDITION,
	    W_BODY
	} p_while;
	enum {
	    F_VARIABLE,
	    F_FROM,
	    F_BY,
	    F_TO,
	    F_BODY
	} p_for;
	enum {
	    C_PROC,
	    C_PARAMETERS
	} p_call;
    },

    P_PTRTYPE = union {
	*PROCTYPE p_procPtr;
	*NODETYPE p_node;
	*IDLISTTYPE p_idList;
	*DECLLISTTYPE p_declList;
	*IFLISTTYPE p_ifList;
	*NODELISTTYPE p_nodeList;
    },

    POSITIONTYPE = struct {
	P_TYPETYPE p_type;
	P_POSITIONTYPE p_position;
	P_PTRTYPE p_ptr;
    };

*IDLISTTYPE IDLISTNULL = pretend(1, *IDLISTTYPE);
*DECLLISTTYPE DECLLISTNULL = pretend(1, *DECLLISTTYPE);
*NODELISTTYPE NODELISTNULL = pretend(1, *NODELISTTYPE);
*IFLISTTYPE IFLISTNULL = pretend(1, *IFLISTTYPE);
*NODETYPE NODENULL = pretend(1, *NODETYPE);

*IDLISTTYPE SymbolTable;

*POSITIONTYPE CurrentPosition;		/* current position in stack */

word
    StartLine,				/* line to start displaying on */
    StopLine,				/* line to stop displaying on */
    CurrentStartLine,			/* line currently at screen top */
    CurrentLine,			/* line current token starts on */
    TargetLine;				/* line current line starts on */

ushort
    NLines,				/* number of lines on screen */
    NCols,				/* number of columns on screen */
    TargetIndent,			/* indent amount of target node */
    CurrentColumn,			/* column current token starts on */
    CurrentLength;			/* length of current edit node */

[TEXTBUFFERLENGTH] char
    TextBuffer,				/* text editing buffer */
    Buffer1,				/* text buffers for strings */
    Buffer2;

bool
    HadEditMessage,			/* had error in visual mode */
    Locating,				/* 'display' is locating a node */
    Headering,				/* 'display' is just "headering" */
    FoundHeader,			/* found the node in headering */
    Displaying;				/* 'display' is actively drawing */

[MAXEDITDEPTH] POSITIONTYPE EditStack;	/* stack of "open" nodes */
